# Author: Martin Matusiak <numerodix@gmail.com>
# Licensed under the GNU Public License, version 3
#
# <desc> Establish state of network/internet connection </desc>
#
# <usage>
# source this file in bash, then run `havenet`
# </usage>


function havenet {
	_platforminit
	case "$PLATFORM" in
		"Linux")
			echo -e "Platform: ${ccyan}$PLATFORM${creset}"
			_run_havenet "$(_toolsinit)";;
		*)
			echo -e "Platform: ${ccyan}$PLATFORM${creset} (local network detection not supported)"
			_run_haveinet "$(_toolsinit)";;
	esac
}

function haveinet {
	_platforminit
	echo -e "Platform: ${ccyan}$PLATFORM${creset}"
	_run_haveinet "$(_toolsinit)"
}

function wifi {
	_platforminit
	echo -e "Platform: ${ccyan}$PLATFORM${creset}"
	_run_wifi
}

function wifiscan {
	local iface="$1";
	_platforminit
	echo -e "Platform: ${ccyan}$PLATFORM${creset}"
	_run_wifiscan "$iface"
}



function _platforminit {
	if $(which uname &>/dev/null); then
		PLATFORM=$(uname 2>/dev/null)
	fi
}

function _toolsinit {
# produce eval'able packet of names mapped to tools
	local missing=

	# init tools
	_route_lookup &>/dev/null
	_ifconfig_lookup &>/dev/null
	_ping_host &>/dev/null
	_nmap_host &>/dev/null

	[[ "$__route" == "missing" ]] && missing="$missing /sbin/route"
	[[ "$__ifconfig" == "missing" ]] && missing="$missing /sbin/ifconfig"
	[[ "$__ping" == "missing" ]] && missing="$missing ping"
	[[ "$__nmap" == "missing" ]] && missing="$missing nmap"

	if [ ${#missing} -gt 0 ]; then
		echo -e "Missing tools: ${cred}${missing}${creset}" 1>&2
	fi
}

function _run_havenet {

	local localranges="127.0.0 169.254 10 172.(1[6-9]|2[0-9]|3[0-1]) 192.168"

	### Scan networks
	
	local networks=
	echo -e "${cyellow} + Scanning for networks...${creset}"

	local route=$(_route_lookup)
	local nets=$(echo "$route" | egrep '^[1-9]' | awk '{ print $1 }')
	for net in $nets; do
		local mask=$(echo "$route" | egrep "^$net" | awk '{ print $3 }' | head -n1)
		local ifaces=$(echo "$route" | egrep "^$net" | awk '{ print $8 }')
		for iface in $ifaces; do
			networks="$networks $net/$mask/$iface"
		done
	done
	networks=$(echo $networks | tr ' ' '\n' | sort | uniq | xargs)
	
	if [[ ${#networks} -eq 0 ]]; then
		echo -e "    ${cred}none found${creset}"
	else
		networks=$(_sort_ips $networks)
		for network in $networks; do
			local chunks=(${network//\// })
			printf "    ${cmagenta}%-10s${creset}" "<${chunks[2]}>"
			printf "  ${cgreen}%-15s${creset}" "${chunks[0]}"
			printf " ${ccyan}%-17s${creset}\n" "/ ${chunks[1]}"
		done
	fi
	
	### Detect ips

	local ips=
	echo -e "${cyellow} + Detecting ips...${creset}"

	local ifconfig=$(_ifconfig_lookup)
	local ifaces=$(echo "$ifconfig" | grep '^[^ ]' | awk '{ print $1 }')
	
	for iface in $ifaces; do
		local data=$(echo "$ifconfig" | sed -e '/./{H;$!d;}' -e "x;/$iface/!d;")

		local ipdata=$(echo "$data" | grep "inet addr:")
		if [[ $? == 0 ]]; then
			local ip=$(echo "$ipdata" | sed "s/.*inet addr:\([0-9.]*\).*$/\1/g" | tr -d ' ')
		fi

		local maskdata=$(echo "$data" | grep "Mask:")
		if [[ $? == 0 ]]; then
			local mask=$(echo "$maskdata" | sed "s/.*Mask:\([0-9.]*\).*$/\1/g" | tr -d ' ')
		fi

		if [[ "$ip" ]] && [[ "$mask" ]]; then
			ips="$ips $ip/$mask/$iface"
		fi
	done

	if [[ ${#ips} -eq 0 ]]; then
		echo -e "    ${cred}none found${creset}"
	else
		local on_lan=1

		ips=$(_sort_ips $ips)
		for ip in $ips; do
			local chunks=(${ip//\// })

			### Test ips for lan
			ip_on_lan=0
			for prefix in $localranges; do
				test=$(echo "${chunks[0]}" | egrep "^$prefix" 2>/dev/null)
				if [[ $? == 0 ]]; then
					ip_on_lan=1
				fi
			done
			on_lan=$(( $on_lan & $ip_on_lan ))

			printf "    ${cmagenta}%-10s${creset}" "<${chunks[2]}>"
			if [[ $ip_on_lan -eq 0 ]]; then  # highlight ip if on wan
				printf "  ${cbgreen}%-15s${creset}" "${chunks[0]}"
			else
				printf "  ${cgreen}%-15s${creset}" "${chunks[0]}"
			fi
			printf " ${ccyan}%-17s${creset}   " "/ ${chunks[1]}"
			_ping_host ${chunks[0]}; echo
		done


		### Detect gateways

		local gateways=
		echo -e "${cyellow} + Detecting gateways...${creset}"
		local gws=$(echo "$route" | grep UG | awk '{ print $2 }')
		for gw in $gws; do
			local iface=$(echo "$route" | egrep "$gw" | awk '{ print $8 }')
			gateways="$gateways $gw/$iface"
		done

		if [[ ${#gateways} -eq 0 ]]; then
			echo -e "    ${cred}none found${creset}"
		else
			gateways=$(_sort_ips $gateways)
			for gateway in $gateways; do
				local chunks=(${gateway//\// })
				printf "    ${cmagenta}%-10s${creset}" "<${chunks[1]}>"
				printf "  ${cgreen}%-15s${creset}   " "${chunks[0]}"
				_ping_host ${chunks[0]}; echo

				for ip in $ips; do
					local ipchunks=(${ip//\// })

					# check if gateway belongs on network of ip
					# 1. network := ip & netmask
					# 2. network' := gw & netmask
					# 3. network == network' ?

					local gwbits=$(_ip_int_to_bits ${chunks[0]})
					local ipbits=$(_ip_int_to_bits ${ipchunks[0]})
					local maskbits=$(_ip_int_to_bits ${ipchunks[1]})

					local netbits=$(( ipbits & maskbits ))
					local netbits2=$(( gwbits & maskbits ))

					if (( netbits == netbits2 )); then
						# highlight external ip
						printf "     %-10s ${cbgreen}%-15s${creset}\n" "ip:" "${ipchunks[0]}"
					fi
				done
			done
		fi

		### Try inet connection if we're on wan or we have a gateway 
		if [[ ! "$on_lan" ]] || [[ ${#gateways} -gt 0 ]]; then
			_run_haveinet
		fi
	fi
}


function _run_haveinet {

	local rootname="b.root-servers.net."
	local rootip="192.228.79.201"
	
	local dnsport="53"

	local inethosts="yahoo.com google.com"
	local inetport="80"

	### Test inet connection

	echo -e "${cyellow} + Testing internet connection...${creset}"
	echo -en "    ${ccyan}$rootname  ${cgreen}$rootip${creset}   "
	_ping_host $rootip; echo

	### Detect dns

	echo -e "${cyellow} + Detecting dns servers...${creset}"
	test=$(cat /etc/resolv.conf 2>/dev/null | grep ^nameserver)
	if [[ $? != 0 ]]; then
		echo -e "    ${cred}none found${creset}"
	else
		local dnss=$(echo "$test" | awk '{ print $2 }')
		dnss=$(_sort_ips $dnss)
		for dns in $dnss; do
			printf "    ${cgreen}%-15s${creset}   " "$dns"
			_ping_host $dns

			echo -en "   "
			_nmap_host "dns" "tcp" $dns $dnsport; echo
		done
	fi

	### Test inet dns

	echo -e "${cyellow} + Testing internet dns...${creset}"

	for inethost in $inethosts; do
		printf "    ${cgreen}%-15s${creset}   " "$inethost"
		_ping_host $inethost

		echo -en "   "
		_nmap_host "http" "tcp" $inethost $inetport; echo
	done

}

function _run_wifi {

	# Check for iwconfig
	if ! $(which /sbin/iwconfig &>/dev/null); then
		echo -e "Not detected: ${cred}/sbin/iwconfig${creset}" 1>&2
		return
	fi

	echo -e "${cyellow} + Detecting wireless interfaces...${creset}"
	local iwconfig=$(/sbin/iwconfig 2>/dev/null)

	# Find wireless interfaces
	local ifaces=$(echo "$iwconfig" | grep '^[^ ]' | awk '{ print $1 }')
	if [[ ${#ifaces} -eq 0 ]]; then
		echo -e "    ${cred}none found${creset}"
		return
	fi

	for iface in $ifaces; do
		local data=$(echo "$iwconfig" | sed -e '/./{H;$!d;}' -e "x;/$iface/!d;")
		local essid=$(echo "$data" | grep ESSID | sed "s|.*ESSID:\"\(.*\)\".*$|\1|g")
		local ap=$(echo "$data" | grep "Access Point" | sed "s|.*Access Point: \([0-9a-fA-F:]*\).*|\1|g")
		local asso=$(echo "$data" | grep "Access Point: Not-Associated" &>/dev/null; echo $?)
		local bitrate=$(echo "$data" | grep "Bit Rate" | sed "s|.*Bit Rate.\([^ ]* [^ ]*\/s\).*$|\1|g")
		local key=$(echo "$data" | grep "Encryption key" | sed "s|.*Encryption key.\([^ ]*\).*$|\1|g" | tr -d '-')
		local quality=$(echo "$data" | grep "Link Quality" | sed "s|.*Link Quality.\([^ ]*\).*$|\1|g")
		local signal=$(echo "$data" | grep "Signal level" | sed "s|.*Signal level.\([^ ]*\).*$|\1|g")
		local noise=$(echo "$data" | grep "Noise level" | sed "s|.*Noise level.\([^ ]*\).*$|\1|g")

		local key_ascii=
		for (( i=0; i<${#key}; i+=2 )); do
			local char=$(printf "\x${key:$i:2}" 2>/dev/null)
			if [[ $(echo "$char" | wc -c) != 2 ]]; then
				key_ascii=
				break
			fi
			key_ascii="$key_ascii$char"
		done


		if [[ -z "$essid" ]];
		then essid=$(printf "${cred}%-17s${creset}" "N/A")
		else essid=$(printf "${cbgreen}%-17s${creset}" "$essid")
		fi

		if [[ -z "$ap" ]];
		then ap=$(printf "${cred}%-17s${creset}" "Not Associated")
		else ap=$(printf "${cgreen}%-17s${creset}" "$ap")
		fi

		if [[ -z "$bitrate" ]];
		then bitrate=$(printf "${cred}%-7s${creset}" "N/A")
		else bitrate=$(printf "${cgreen}%-7s${creset}" "$bitrate")
		fi

		if [[ -z "$quality" ]] || [[ "$quality" == "0" ]];
		then quality=$(printf "${cred}%-7s${creset}" "N/A")
		else quality=$(printf "${cgreen}%-7s${creset}" "$quality")
		fi

		if [[ -z "$signal" ]] || [[ "$signal" == "0" ]];
		then signal=$(printf "${cred}%-8s${creset}" "N/A")
		else signal=$(printf "${cgreen}%-8s${creset}" "$signal dBm")
		fi

		if [[ -z "$noise" ]] || [[ "$noise" == "0" ]];
		then noise=$(printf "${cred}%-8s${creset}" "N/A")
		else noise=$(printf "${cgreen}%-8s${creset}" "$noise dBm")
		fi

		if [[ -z "$key" ]]; then
			key="not detected"
			if [[ $(whoami) != "root" ]]; then
				key="$key (not root)"
			fi
			key="${cred}$key${creset}"
		elif [[ "$key" == "off" ]]; then
			key="${cgreen}none${creset}"
		else
			key="${cgreen}$key${creset}"
		fi

		printf "    ${cmagenta}%-10s${creset}" "<$iface>"
		printf "    ${cyellow}<essid>${creset} $essid"
		printf "    ${cyellow}<ap>${creset} $ap\n"
		printf "     quality: $quality"
		printf "  bitrate: $bitrate"
		printf "   signal: $signal"
		printf "  noise: $noise\n"

		printf "     key: $key"
		if [[ -n "$key_ascii" ]]; then
			printf "  '${cgreen}$key_ascii${creset}'"
		fi
		echo
	done
}

function _run_wifiscan {
	local useiface="$1";

	# Check for iwconfig, iwlist
	local missing=
	if ! $(which /sbin/iwconfig &>/dev/null); then
		missing="$missing /sbin/iwconfig"
	fi
	if ! $(which /sbin/iwlist &>/dev/null); then
		missing="$missing /sbin/iwlist"
	fi

	if [[ -n "$missing" ]]; then
		echo -e "Not detected: ${cred}$missing${creset}" 1>&2
		return
	fi

	# Use declared interface or detect interfaces
	if [[ -n "$useiface" ]]; then
		local ifaces=$useiface
	else
		echo -e "${cyellow} + Detecting wireless interfaces...${creset}"
		local iwconfig=$(/sbin/iwconfig 2>/dev/null)

		# Find wireless interfaces
		local ifaces=$(echo "$iwconfig" | grep '^[^ ]' | awk '{ print $1 }')
		if [[ ${#ifaces} -eq 0 ]]; then
			echo -e "    ${cred}none found${creset}"
			return
		else
			for iface in $ifaces; do
				printf "    ${cmagenta}%-10s${creset}\n" "<$iface>"
			done
		fi
	fi

	for iface in $ifaces; do
		# Scan for access points
		echo -e "${cyellow} + Scanning for access points on <$iface>...${creset}"

		local iwlist=$(/sbin/iwlist $iface scanning | sed "s|Cell|\nCell|g")
		local cells=$(echo "$iwlist" | grep Cell | sed "s|.*Cell \([0-9]*\).*$|\1|g" | xargs)

		if [[ ${#cells} -eq 0 ]]; then
			echo -e "    ${cred}none found${creset}"
			continue
		fi

		local _aps=
		for cell in $cells; do
			local data=$(echo "$iwlist" | sed -n "/Cell $cell/,/^[ ]*$/p")

			local essid=$(echo "$data" | grep ESSID | sed "s|.*ESSID:\"\(.*\)\".*$|\1|g")
			local ap=$(echo "$data" | grep "Address" | sed "s|.*Address: \([0-9a-fA-F:]*\).*|\1|g")
			local asso=$(echo "$data" | grep "Access Point: Not-Associated" &>/dev/null; echo $?)
			local bitrate=$(echo "$data" | xargs | sed "s|.*Bit Rates:\(\([0-9.]* [Mbs;/]*\)*\).*|\1|g" | tr ' ' '\n' | grep ^[0-9] | tail -n1)
			local key=$(echo "$data" | grep "Encryption key" | sed "s|.*Encryption key.\([^ ]*\).*$|\1|g" | tr -d '-')
			local quality=$(echo "$data" | grep "Quality" | sed "s|.*Quality.\([^ ]*\).*$|\1|g")
			local signal=$(echo "$data" | grep "Signal level" | sed "s|.*Signal level.\([^ ]*\).*$|\1|g")
			local noise=$(echo "$data" | grep "Noise level" | sed "s|.*Noise level.\([^ ]*\).*$|\1|g")

			_aps="$_aps\n$quality===essid:$essid===ap:$ap===key:$key===bitrate:$bitrate===signal:$signal===noise:$noise"
		done

		# Sort and print access points
		_aps=$(echo -e "$_aps" | sort -n)

		echo "$_aps" | while read _ap; do
			[[ -z "$_ap" ]] && continue

			_ap=$(echo "$_ap" | sed "s|===|\n|g")
			local essid=$(echo "$_ap" | grep "essid" | sed "s|^.*essid:\(.*\).*$|\1|g")
			local ap=$(echo "$_ap" | grep "ap" | sed "s|^.*ap:\(.*\).*$|\1|g")
			local key=$(echo "$_ap" | grep "key" | sed "s|^.*key:\(.*\).*$|\1|g")
			local bitrate=$(echo "$_ap" | grep "bitrate" | sed "s|^.*bitrate:\(.*\).*$|\1|g")
			local quality=$(echo "$_ap" | head -n1)
			local signal=$(echo "$_ap" | grep "signal" | sed "s|^.*signal:\(.*\).*$|\1|g")
			local noise=$(echo "$_ap" | grep "noise" | sed "s|^.*noise:\(.*\).*$|\1|g")

			if [[ -z "$essid" ]];
			then essid=$(printf "${cred}%-17s${creset}" "N/A")
			else essid=$(printf "${cbgreen}%-17s${creset}" "$essid")
			fi

			if [[ -z "$ap" ]];
			then ap=$(printf "${cred}%-17s${creset}" "N/A")
			else ap=$(printf "${cgreen}%-17s${creset}" "$ap")
			fi

			if [[ -z "$bitrate" ]];
			then bitrate=$(printf "${cred}%-7s${creset}" "N/A")
			else bitrate=$(printf "${cgreen}%-7s${creset}" "$bitrate Mb/s")
			fi

			if [[ -z "$quality" ]] || [[ "$quality" == "0" ]];
			then quality=$(printf "${cred}%-7s${creset}" "N/A")
			else quality=$(printf "${cgreen}%-7s${creset}" "$quality")
			fi

			if [[ -z "$signal" ]] || [[ "$signal" == "0" ]];
			then signal=$(printf "${cred}%-8s${creset}" "N/A")
			else signal=$(printf "${cgreen}%-8s${creset}" "$signal dBm")
			fi

			if [[ -z "$noise" ]] || [[ "$noise" == "0" ]];
			then noise=$(printf "${cred}%-8s${creset}" "N/A")
			else noise=$(printf "${cgreen}%-8s${creset}" "$noise dBm")
			fi

			if [[ -z "$key" ]]; then
				key="${cred}N/A${creset}"
			elif [[ "$key" == "on" ]]; then
				key="${ccyan}$key${creset}"
			else
				key="${cgreen}$key${creset}"
			fi

			printf "    ${cyellow}<essid>${creset} $essid"
			printf "    ${cyellow}<ap>${creset} $ap"
			printf "    ${cyellow}<key>${creset} $key\n"
			printf "      quality: $quality"
			printf "  bitrate: $bitrate"
			printf "   signal: $signal"
			printf "  noise: $noise\n"
		done
	done
}


function _route_lookup {
	if [[ -z "$__route" ]]; then
		__route="missing"
		if which /sbin/route &>/dev/null; then
			__route="/sbin/route -n"
		fi
	fi
	
	if [[ "$__route" != "missing" ]]; then
		local res=$($__route 2>/dev/null)
		echo "$res"
	fi
}

function _ifconfig_lookup {
	if [[ -z "$__ifconfig" ]]; then
		__ifconfig="missing"
		if which /sbin/ifconfig &>/dev/null; then
			__ifconfig="/sbin/ifconfig -a"
		fi
	fi
	
	if [[ "$__ifconfig" != "missing" ]]; then
		local res=$($__ifconfig 2>/dev/null)
		echo "$res"
	fi
}

function _ping_host {
	local host="$1";shift;

	if [[ -z "$__ping" ]]; then
		__ping="missing"
		if which ping &>/dev/null; then
			local res=$(ping -h 2>&1)
			if echo "$res" | grep '[[]-c count[]]' &>/dev/null; then
				__ping="ping -c1 -W2 %s"  # linux ping
				if echo "$res" | grep '[[]-t timeout[]]' &>/dev/null; then
					__ping="ping -c1 -t2 %s"  # freebsd ping
				fi
			else
				__ping="ping %s 56 1"  # fping (solaris, cygwin)
			fi
		fi
	fi
	
	if [[ "$__ping" != "missing" ]] && [[ -n "$host" ]]; then
		echo -en "ping: "
		local cmd=$(printf "$__ping" "$host")
		local res=$($cmd 2>/dev/null)
		local t=
		if echo "$res" | grep "time=" &>/dev/null; then
			t="$(echo "$res" | grep "time=" | sed "s|.*time=\([0-9.]*\).*$|\1|g") ms"
		elif echo "$res" | grep "alive" &>/dev/null; then
			t="alive"
		fi
		if [[ "$t" ]]; then
			echo -en "${cgreen}$t${creset}"
		else
			echo -en "${cred}failed${creset}"
		fi
	fi
}

function _nmap_host {
	local label="$1";shift;
	local proto="$1";shift;
	local host="$1";shift;
	local port="$1";shift;
	local flags="$1";shift;

	if [[ -z "$__nmap" ]]; then
		__nmap="missing"
		if which nmap &>/dev/null; then
			__nmap="nmap"
		fi
	fi
	
	if [[ "$__nmap" != "missing" ]] && [[ "$host" ]]; then
		local udpflags=""
		if [[ "$port" == "53" ]] && [[ `whoami` = "root" ]]; then
			proto="udp"; udpflags="-sU -PN"
		fi

		printf "%s: " "$label/$proto"

		local cmd=$(printf "%s -p %s %s %s %s" "$__nmap" "$port" "$host" "$flags" "$udpflags")
		local res=$($cmd 2>/dev/null)
		local t=
		if echo "$res" | grep "$port/$proto" | grep "open" &>/dev/null; then
			t=$(echo "$res" | grep "scanned in" | sed "s/^.*in \([0-9.]*\) seconds.*$/\1/g")
			t=$(echo "scale=0; $t*1000/1" | bc)
		fi
		if [[ "$t" ]]; then
			echo -en "${cgreen}$t ms${creset}"
		else
			echo -en "${cred}failed${creset}"
		fi
	fi
}


function _sort_ips {
# sort ips numerically
	local ips="$@"

	# insert leading zeros
	local ips_p=
	for ip in $ips; do
		local suffix=${ip#*/}
		ip=${ip%%/*}
		[[ "$suffix" == "$ip" ]] && suffix= || suffix="/$suffix"

		local octs=${ip//./ }
		local octs_p=
		for oct in $octs; do
			octs_p="${octs_p}$(printf ".%03d" $oct)"
		done
		ips_p="${ips_p} ${octs_p:1}$suffix"
	done

	# sort lexicographically
	ips_p=$(echo $ips_p | tr ' ' '\n' | sort)

	# remove leading zeros
	ips=
	for ip in $ips_p; do
		local suffix=${ip#*/}
		ip=${ip%%/*}
		[[ "$suffix" == "$ip" ]] && suffix= || suffix="/$suffix"

		octs_p=${ip//./ }
		octs=
		for oct in $octs_p; do
			oct=$(echo $oct | sed "s/^0\{1,2\}//g")
			octs="${octs}.$oct"
		done
		ips="${ips} ${octs:1}$suffix"
	done

	echo $ips
}

function _ip_int_to_bits {
	local ip="$@"
	local bits=0
	local i=3
	local fields=${ip//./ }
	for field in $fields; do
		bits=$(( bits + (field << (8 * i)) ))
		i=$(( i - 1 ))
	done
	echo $bits
}

function _ip_bits_to_int {
	local bits="$@"
	local ip=
	for (( i=3; i>=0; i-- )); do
		head=$(( bits >> (8 * i) ))
		bits=$(( bits & ~(255 << (8 * i)) ))
		ip=$ip.$head
	done
	echo ${ip:1}
}
